#! /usr/bin/env python


__doc__ = """CallHome.py -- CallHome text tools
"""

__version__ = "$Revision:  $"



from UserList import UserList
from UserDict import UserDict

import string, os

import xml.sax,xml.sax.handler


def validGame(gamein):

    if gamein is None: return None

    import re,string
    game = string.split(gamein,'^')

    major = game[0]
    minor = game[1:]

    if major=='seek' and minor[0]=='conf': # Control acts is a special class
        major='seek^conf'
        del minor[0]

    if major not in ['info','quest','opinion','direct','express','filler','seek^conf','commit']:
        raise 'Invalid major game tag in %s'%gamein,str(major)

    for m in minor:
        if m not in ['elab','st','spec','inc','aband','wk','sup']:
            raise 'Invalid minor game tag in %s'%gamein,str(m)
    return gamein


def validDA(dain,saact):
    import re,string
    da = string.split(dain,'^')

    major = da[0]
    minor = da[1:]

    if major=='ca': # Control acts is a special class
        major='ca^'+minor[0]
        del minor[0]

    # Check if minor DA is sorted
    minor2 = minor[:]
    minor.sort()
    if minor!=minor2: # unsorted
        raise 'UNSORTED',dain    
    dain = string.join([major]+minor,'^')


    if da[0]=='s':
        # remove semantic feature dimensions
        minor = filter(lambda x:x not in ['me','hyp','val+','val-','obl','fut','cert+','cert-','jk'],minor)

    if da[0] in ['qw','qy','qh','qr','qd','qo']:
        # remove communication problems and 'intent but not form of questions'
        minor = filter(lambda x:x not in ['c','d','g'],minor)

    # remove all other tags (are all orthogonal)    
    minor = filter(lambda x:x not in ['t1','t3','ins','is','q','m','2','r','e','cor','fp','fc','g'],minor)        


    if major not in ["%","aa","ar","atd","","b","b!","b+","b-","bc","bh","br","ca^com","ca^dir","ca^off","ca^prq","ca^sug","fa","fc","fe","ff","fp","ft","h","l","na","nd","ng","nn","no","ny","p","qh","qo","qr","qw","qy","s","x"]:
        raise 'Invalid major dialogue act tag in %s'%dain,str(major)

    if minor:
        print 'Invalid minor dialogue act tag in %s'%dain,str(minor)
        for wrong in minor:
            invalid[wrong].append(saact)

        raise 'Invalid minor dialogue act tag in %s'%dain,str(minor)

    return dain







# Convert ISO encoding into standard strings

def enc(x):
    import string
    if x is None: return None
    return string.join(map(lambda x:chr(ord(x)),x),'')
    return ret




# The conversation, genre and dialogue act are encoded as XML tags.
# Checking is done by callig the constructor which does verify the arguments at least
# More checking can be done when the discourse is complete (to be done, implemented in other module)

# The __str__ function is arranged to generate my old file format for cross-checking

class CONVERSATION(UserList):
    def __init__(self,id):
        self.id = int(id)
        UserList.__init__(self)

    def __str__(self):
        import string
        return string.join(map(str,self),'\n')

class GENRE(UserList):
    activity = ['Inform','Inquire','Planning','Convince','Negotiate','Gossip','Argue','Conver-mgnt','Greet-close','Instruct']
    topic    = ['Travel','Job','Free tel. call','Education','Career Planning','Politics','Foreign Country','Money','Health','Well Being']    

    actp,topp = 'Activity:','Topic:'

    def __init__(self,id,Who_or_What,Important_news,Topic_Other,**kargs):
        self.id = enc(id)

        self.wow,self.topicOther,self.news = map(enc,[Who_or_What,Topic_Other,Important_news])

        self.activity = {}
        self.topic    = {}

        for k,v in kargs.items():
            import string
            k = string.replace(k,"Activity_","Activity:")
            k = string.replace(k,"Topic_","Topic:")
            k = string.replace(k,"tel_","tel. ")
            k = string.replace(k,"_mgnt","-mgnt")
            k = string.replace(k,"_close","-close")
            k = string.replace(k,"_"," ")
            if k[:len(GENRE.actp)]==GENRE.actp and k[len(GENRE.actp):] in GENRE.activity:
                v = int(v)
                if v not in range(6): raise 'Invalid range',v                
                self.activity[k[len(GENRE.actp):]] = v
            elif k[:len(GENRE.topp)]==GENRE.topp and k[len(GENRE.topp):] in GENRE.topic:
                v = int(v)
                if v not in range(6): raise 'Invalid range',v                
                self.topic[k[len(GENRE.topp):]] = v
            else:
                raise TypeError,'unexpected keyword argument: %s'%k

        # Check if we picked up all activities / topics

        for t in ['activity','topic']:

            missing = []
            l,arr = getattr(GENRE,t),getattr(self,t)
            for e in l:
                if not arr.has_key(e):
                    missing.append('%s of %s'%(e,t))
            if missing:
                raise TypeError,'not enough arguments; missing'%(e,string.join(missing))
        
        UserList.__init__(self)
    def __str__(self):
        import string

        ret = ['#Topic Boundary','#Genre: %s'%self.id]

        for e in GENRE.activity:
            ret.append('#Activity:%s\t%d'%(e,self.activity[e]))

        ret.append('#Who or What:\t%s'%self.wow)

        for e in GENRE.topic:
            ret.append('#Topic:%s\t%d'%(e,self.topic[e]))

        ret.append('#Topic Other:\t%s'%self.topicOther)
        ret.append('#Important news:\t%s'%(self.news))

        ret[len(ret):] = map(str,self)

        return string.join(ret,'\n')


def validGame(gamein):

    if gamein is None: return None

    import re,string
    game = string.split(gamein,'^')

    major = game[0]
    minor = game[1:]

    if major=='seek' and minor[0]=='conf': # Control acts is a special class
        major='seek^conf'
        del minor[0]

    if major not in ['info','quest','opinion','direct','express','filler','seek^conf','commit']:
        raise 'Invalid major game tag in %s'%gamein,str(major)

    for m in minor:
        if m not in ['elab','st','spec','inc','aband','wk','sup']:
            raise 'Invalid minor game tag in %s'%gamein,str(m)
    return gamein


class DA(UserList):
    def __init__(self,id,channel,initiative,game=None,willbenested=None,amnested=None,translation=None):
        self.id,self.initiative,self.game,self.willbenested,self.amnested,self.translation,self.channel=map(enc,[id,initiative,game,willbenested,amnested,translation,channel])
        validGame(self.game)
        UserList.__init__(self)
        validDA(self.id,self)

    def __getattr__(self,attr):
        import string
        if attr=='string':
            return enc(string.join(self.data,''))
        elif attr=='words':
            return string.split(self.string)
        else:
            return UserList.__getattr__(self,attr)
            

    def __str__(self):

        initiative = self.initiative
        if self.amnested: initiative = initiative+'^n'

        if self.game is None:
            game = '<%s>'%initiative
        elif self.willbenested:
            game = '%s^n <%s>'%(self.game,initiative)
        else:
            game = '%s <%s>'%(self.game,initiative)

        ret = string.rjust('%s:%s %s: '%(game,self.id,self.channel),25)+self.string
	if self.translation is not None:
	    ret = ret + '\n' + string.rjust('@ ',25) + enc(self.translation)
        return ret


# XML-parser ContentHandler
# For each type that is recognized an attribute is available as long as the type is still "open"
# Lower level types just append to the upper level type (DAs are appended to GENRE, GENRE to CONVERSATION, CONVERSATION to self.list)
# self.list is appended a conversation


class ContentHandler(xml.sax.handler.ContentHandler):

    paras = ['CONVERSATION','GENRE','DA']

    def __init__(self,list):
        self.list = list
        xml.sax.handler.ContentHandler.__init__(self)
        
    def startElement(self, name, attrs):
        if name in ContentHandler.paras:
            for i in range(len(ContentHandler.paras)):
                if name==ContentHandler.paras[i]: break

            import __main__

            # First check that it's not there (ERROR OTHERWISE)

            if hasattr(self,name): raise 'name defined',name

            # Create the object -- consistency is checked in the constructor
            at = {}
            for k,v in attrs.items():
                k = str(k)
                if name=='GENRE':
                    k = string.replace(k,"-","_")                    
                at[str(k)]=v
            setattr(self,name,apply(__main__.__dict__[name],(),at))
            
            # Append to the next higher level -- implicitly checking
            getattr(self,(['list']+ContentHandler.paras)[i]).append(getattr(self,name))
        else:
            raise 'Illegal element',name+str(attrs.items())

    def endElement(self, name):
        if name in ContentHandler.paras:  del self.__dict__[name]
        else:                             raise 'END',name

    def characters(self, content):
        import string

        if hasattr(self,'DA'):       self.DA.append(content)
        elif string.strip(content):  raise 'CONTENT "%s"'%content

    def ignorableWhitespace(self, whitespace): raise 'WHITE',whitespace
    def skippedEntity(self, name):             raise 'Skipped',name


# Checks whether a parsed dialogue conforms with the original as in the LDC CallHome Spanish distribution
# If the channels are different, fail
# Otherwise write diff files in diff0 and diff1
# There are actually some diffs due to the fact that in the original "{laugh}hello" occurs and these two words
# go into different dialogue acts although not being whitespace seperated.
# This is the case for 8 cases in 7 dialogues.


NoOriginal = 'Not in database'


def checkoriginal(dialogue,base):
    import string,re,os,glob

    channelsDialogue,channelsFile = {},{}
    matchchannel = re.compile(r"([AB]\d*):+$")

    # Read all the information from each channel in the dialogue
    # Ignore the [pauses] tokens
    
    for genre in dialogue:
        for da in genre:
            l = channelsDialogue.setdefault(da.channel,[])
            data = filter(lambda x:x!='[PAUSE]',da.words)
            l[len(l):] = data

    # Open the file in the database and loop over all of its lines

    for db in ['devtest','evltest','train']:
        f = os.path.join(base,db,'sp_%04d.txt'%conversation.id)
        if os.path.isfile(f):
            for line in open(glob.glob(os.path.join(base,'*/sp_%04d.txt'%conversation.id))[0]).readlines():
                start,end,channel,line = string.split(line,None,3)
                channel = matchchannel.match(channel).group(1)
                l = channelsFile.setdefault(channel,[])
                l[len(l):] = string.split(line)
            break
        if db=='train': raise NoOriginal,'sp_%04d.txt'%conversation.id

    cd,cf = channelsDialogue.keys(),channelsFile.keys()
    cd.sort() ; cf.sort()
    
    if cd!=cf: raise 'Channels are different',str(cd)+' '+str(cf)

    for channel in cd:
        import string
        
        cf,cd = string.join(channelsFile[channel],' '),string.join(channelsDialogue[channel],' ')
        if enc(cf)!=enc(cd):
            try: os.mkdir('diff0')
            except: pass
            try: os.mkdir('diff1')
            except: pass
            
            open('diff0/%ld_%s'%(dialogue.id,channel),'w').write(string.lower(string.join(map(enc,channelsFile[channel]),'\n')))
            open('diff1/%ld_%s'%(dialogue.id,channel),'w').write(string.lower(string.join(map(enc,channelsDialogue[channel]),'\n')))
            if string.join(channelsFile[channel],'') == string.join(channelsDialogue[channel],''):                
                print 'Channels are the same but some whitespaces are different %s %s'%(dialogue.id,channel)
            else:
                print 'Channels are different %s %s'%(dialogue.id,channel)



def createDaInventory(da):

    daTypes = {}
    dialogueDA = {}

    for dialogue in da:

        das = []
        for topic in dialogue:
            for sa in topic:
                das.append((dialogue.id,len(das),sa))
                    
        dialogueDA[dialogue.id] = das

    for das in dialogueDA.values():
        for sa in das:
            if not daTypes.has_key(sa[2].id): daTypes[sa[2].id] = []
            daTypes[sa[2].id].append(sa)

    def pr(s):
        s = str(s)
        import re,string
        #s = re.sub(r"\^",r"\^{ }",s)
        #s = re.sub(r"([<>])",r"$\1$",s)
        s = re.sub(r"([<>])",r"",s)
        #s = re.sub(r"([%_])",r"\\\1",s)
        return string.strip(s)

    print """
<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">
<html>
<head>
<title>Dialogue Act Crossreference</title>

<h1>Dialogue Acts
</h1><hr><center>

"""

    daTypes = daTypes.items() ; daTypes.sort()

    for datype,das in daTypes:
        print "<b><a href=\"#DA-%s\"> %s </a></b> |"%(pr(datype),pr(datype))
    print "</center>"


    for datype,das in daTypes:
        print "<h2> <a name=\"DA-%s\"> %s</a> </h2>\n\n"%(pr(datype),pr(datype))

        ls = []
        for  dialogue,i,da in das:
            ls.append((pr(da),"<a href=\"#%ld:%ld\"> %s </a> <Br>"%(dialogue,i,pr(da))))
        ls.sort(lambda x,y:cmp(len(x[0]),len(y[0])) or cmp(x[0],y[0]))
        for e in ls:
            print e[1]

    print """

<h1>Dialogues</h1>
"""

    for dialogueid,das in dialogueDA.items():
        print "<h2> <a name=\"DIALOGUE-%s\"> %s</a> </h2>\n\n"%(dialogueid,dialogueid)
        for  dialogue,i,da in das:
            print "<a name=\"%ld:%ld\"> %s </a> <Br>"%(dialogue,i,pr(da))
        


class Count:
    def __init__(self):
        self.args = {}

    """ Args.Count -- counting with an array that defaults to zero (less code) """

    def __getitem__(self,key):
        return self.args.get(key,0)

    def __setitem__(self,key,value):
        self.args[key]=value

    def __getattr__(self,attr):
        return getattr(self.args,attr)

    def inc(self,key):
	ret = self[key]+1
        self[key] = ret
	return ret

    def pr(self):
	import string
	from operator import add

	li = self.items()
	li.sort(lambda x,y: cmp(y[1],x[1]))

	sum  = reduce(add,self.values(),0)
	if sum==0: return None
	
        lsum = 0
        for i in range(len(li)):
            key,count = li[i]
            lsum = lsum+count
            print '%-70s %ld'%(str(key),count)

if __name__ == '__main__':

    import sys,os,re

    ret = []
    nooriginal = []

    da,game,activity,wow = Count(),Count(),Count(),Count()


    for e in os.listdir('data'):
        m = re.match(r"sp_\d{4}.xml$",e)
        if m:
            sys.stderr.write("%%Processing %s\n"%e)
            if 0: # SAX keeps all files open (why ??)
                xml.sax.parse('data/%s'%e,ContentHandler(ret))
            else: # Workaround: Open the files and parse the strings ;-)
                f = string.join(open('data/%s'%e).readlines(),"") ; xml.sax.parseString(f,ContentHandler(ret))
            conversation = ret[-1]
            try:
                checkoriginal(conversation,'CallHome-Spanish-Original')
            except(NoOriginal):
                nooriginal.append(conversation.id)
                

            for a in conversation:
                activity.inc(a.id)

                wow.inc(re.search('.*\*(\S*?)\*',a.wow).group(1))
                
                for d in a:
                    da.inc(d.id)
                    if d.game is not None:
                        game.inc(d.game)


    if nooriginal:
        sys.stderr.write('In order to check the data against the original distribution it must be provided in \'CallHome-Spanish-Original\'\n')
        sys.stderr.write('The following conversations have not been found:\n')
        sys.stderr.write(string.join(map(str,nooriginal))+'\n')

    if 0: # If you want to create count tables
        print 'Activities'
        print
        activity.pr()
        print

        print 'Who or What major label'
        print
        wow.pr()
        print

        print 'Dialogue acts'
        print
        da.pr()
        print

        print 'Game'
        print
        game.pr()
        print
    

    if 0: # If you want to create the dialogue act cross-reference
        createDaInventory(ret)
    
    

